#pragma warning(disable: 4786) // shut up

/*
 * Example of a Win32 OpenGL program.
 * The OpenGL code is the same as that used in
 * the X Window System sample
 */

#include <vector>
#include <set>
#include <deque>
#include <string>
#include <algorithm>

#include <crtdbg.h>
#include <stdio.h>
#include <windows.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "dxfout.h"
#include "vector.h"

namespace std {
	template<class T>
	inline const T& min(const T& x, const T& y) {
		return _cpp_min(x, y);
	}

	template<class T>
	inline const T& max(const T& x, const T& y) {
		return _cpp_max(x, y);
	}
};

///////////////////////////////////////////////////////////////////////////

/* Windows globals, defines, and prototypes */
CHAR szAppName[]="Aeris";
const char g_szError[]="Aeris Error";
HWND  g_hwnd;
HDC   ghDC;
HGLRC ghRC;

#define SWAPBUFFERS SwapBuffers(ghDC)
#define BLACK_INDEX     0
#define RED_INDEX       13
#define GREEN_INDEX     14
#define BLUE_INDEX      16
#define WIDTH           300
#define HEIGHT          200

LONG WINAPI MainWndProc (HWND, UINT, WPARAM, LPARAM);
BOOL bSetupPixelFormat(HDC);

GLvoid resize(GLsizei, GLsizei);
GLvoid initializeGL(GLsizei, GLsizei);
GLvoid drawScene();

void LoadTextures(std::vector<GLuint>& textures, const struct FF7PackEntry& ent, int& w2, int& h2);

///////////////////////////////

int newidx = 0;

struct FF7Point {
	short x, y, z, w;
};

struct FF7Poly {
	int type;
	int pal;
	short idx[4];
	unsigned char c[4][4];
	int uv[3][2];
};

struct FF7PackEntry {
	const void *p;
	int len;
};

struct FF7BoneEntry {
	const void *p;
	int len;
	int parent_bone;
	int length;
};

static inline double degreesToRadians(double d) {
	return d * (1.0 / 57.295779513082320876798154814105);
}

static inline double blipsToRadians(double d) {
	return d * (3.1415926535897932 / 0x800);
}

class FF7Bone {
public:
	FF7Bone();
	~FF7Bone() throw();

	const std::string& getName() { return mName; }
	void setName(const char *name) { mName = name; }
	void readFrom(const FF7BoneEntry& e) { readFrom((const char *)e.p, e.len); }
	void readFrom(const char *src, int length);
	void draw(const GLuint *textures, bool bTranslucent, double u_factor, double v_factor);
	void drawWireframe();
	void drawShadow();
	void writeDXF(const char *pszFile);

	void writeDXFFromOpenGLTransform(DXF3DOut&);

	void attachTo(FF7Bone& parent, double length);
	void setLength(double l) { mLength = l; }
	void setOrientation(const Quaternion& q) { mOrientation = q; }
	FF7Bone& rotateX(double d) { mOrientation.RotateX(degreesToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateY(double d) { mOrientation.RotateY(degreesToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateZ(double d) { mOrientation.RotateZ(degreesToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateX2(double d) { mOrientation.RotateX(blipsToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateY2(double d) { mOrientation.RotateY(blipsToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateZ2(double d) { mOrientation.RotateZ(blipsToRadians(d)); mOrientation.Normalize(); return *this; }
	FF7Bone& rotateXYZ(int x, int y, int z) {
		return (*this)
			.rotateY2(y)
			.rotateX2(x)
			.rotateZ2(z)
			;
	}
	void rotate(const Quaternion& q) { mOrientation = q * mOrientation; mOrientation.Normalize(); }

	void enumerate(std::vector<FF7Bone *>& v);
	void setupTransform();
	Quaternion getPreOrientation();
	Quaternion getOrientation();
	Quaternion getLocalOrientation() { return mOrientation; }
	void setLocalOrientation(const Quaternion& q) { mOrientation = q; }
	Vector3 getPosition();
	FF7Bone *getParent() { return mpParentNode; }
	double getLength() { return mLength; }
	void setScale(double s) { mScale=s; }
	double getScale() const { return mScale; }

	void writeTo(FILE *f);

private:
	void PolyVert(const FF7Poly& p, int v);
	void PolyVertColor(const FF7Poly& p, int v);
	void PolyVertColorTrans(const FF7Poly& p, int v);
	void _setupTransform(bool);

	std::vector<FF7Point> mPoints;
	std::vector<FF7Poly> mPolygons;
	std::set<FF7Bone *> mChildren;
	std::string mName;

	FF7Bone *mpParentNode;
	double mLength;
	double mScale;
	Quaternion mOrientation;
};

FF7Bone::FF7Bone() : mpParentNode(NULL), mOrientation(Vector3(0.,0.,0.), 1.), mScale(1.0) {
}

FF7Bone::~FF7Bone() throw() {
}

void FF7Bone::readFrom(const char *src, int length) {
	mPoints.clear();
	mPolygons.clear();

	const char *srclimit = src + length;

	// First four bytes are the length of the vertex pool.

	int vplen = *(unsigned *)src;
	src += 4;

	mPoints.resize(vplen);
	memcpy(&mPoints[0], src, vplen);

	src += vplen;

	// Read in polygons until we hit the end of the array.

	while(src < srclimit) {
		unsigned short count = *(short *)src;
		int last_type = 0;

		src += 4;	// don't know what the other two bytes are for

		for(int p=0; p<count; ++p) {
			FF7Poly poly;

			// Read four 16-bit indices.

			memcpy(poly.idx, src, 8);

			_RPT1(0, "[%3d] ", mPolygons.size());
			_RPT4(0, "Indices: %04x %04x %04x %04x\n", poly.idx[0], poly.idx[1], poly.idx[2], poly.idx[3]);

			// The next four bytes are almost always the first vertex
			// color, but we won't know for sure until we read byte 11
			// of the polygon.
			//
			//	0x30 = (20 bytes) three indices + three colors (indexed triangle)
			//	0x32 = (20 bytes) ?
			//	0x38 = (24 bytes) four indices + four colors (indexed tristrip with 2 polygons)
			//	0x78 = (16 bytes) ?

			last_type = poly.type = src[11];

			switch(poly.type) {
			case 0x30:
			case 0x32:
				memcpy(poly.c, src+8, 12);
				src += 20;
				break;
			case 0x38:
				memcpy(poly.c, src+8, 16);
				src += 24;
				break;
			case 0x78:
				poly.pal = (unsigned char)src[10];

				// decode UVs

				poly.uv[0][0] = src[ 8] & 0xff;
				poly.uv[0][1] = src[ 9] & 0xff;
				poly.uv[1][0] = src[12] & 0xff;
				poly.uv[1][1] = src[13] & 0xff;
				poly.uv[2][0] = src[14] & 0xff;
				poly.uv[2][1] = src[15] & 0xff;

				src += 16;
				break;
			default:
				__asm int 3
			}

			mPolygons.push_back(poly);
		}

		_RPT1(0,"Polygons: %d\n", mPolygons.size());
	}
}

void FF7Bone::draw(const GLuint *texture, bool bDrawTranslucent, double u_factor, double v_factor) {
	std::vector<FF7Poly>::iterator it = mPolygons.begin(), itEnd = mPolygons.end();

	glPushMatrix();

	double m[16];

	mOrientation.ToMatrix(m);

	glMultMatrixd(m);
	glPushMatrix();
	glScaled(mScale, mScale, mScale);

	for(; it!=itEnd; ++it) {
		FF7Poly& p = *it;

		if (bDrawTranslucent) {
			switch(p.type) {
			case 0x78:
				{
					int u=0,v=0;

					glBindTexture(GL_TEXTURE_2D, texture[p.pal >> 6]);
						
					glEnable(GL_TEXTURE_2D);
					glEnable(GL_ALPHA_TEST);
					glEnable(GL_BLEND);
					glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
					glColor4f(1,1,1,1);
					glBegin(GL_TRIANGLES);
					glTexCoord2d((p.uv[0][0] + u) * u_factor, (p.uv[0][1]+v) * v_factor);
					PolyVert(p, 0);
					glTexCoord2d((p.uv[1][0] + u) * u_factor, (p.uv[1][1]+v) * v_factor);
					PolyVert(p, 1);
					glTexCoord2d((p.uv[2][0] + u) * u_factor, (p.uv[2][1]+v) * v_factor);
					PolyVert(p, 2);
					glEnd();
					glDisable(GL_BLEND);
					glDisable(GL_ALPHA_TEST);
					glDisable(GL_TEXTURE_2D);
				}
				break;
			}
		} else {
			switch(p.type) {
			case 0x30:
				glBegin(GL_TRIANGLES);
				PolyVertColor(p, 0);
				PolyVertColor(p, 1);
				PolyVertColor(p, 2);
				glEnd();
				break;

			case 0x32:
				glEnable(GL_BLEND);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				glBegin(GL_TRIANGLES);
				PolyVertColorTrans(p, 0);
				PolyVertColorTrans(p, 1);
				PolyVertColorTrans(p, 2);
				glEnd();
				glDisable(GL_BLEND);
				break;

			case 0x38:
				glBegin(GL_TRIANGLE_STRIP);
				PolyVertColor(p, 0);
				PolyVertColor(p, 1);
				PolyVertColor(p, 2);
				PolyVertColor(p, 3);
				glEnd();
				break;
			}
		}
	}
	glPopMatrix();

	glTranslated(0, 0, mLength * mScale);

	std::set<FF7Bone *>::iterator itChildren = mChildren.begin();

	for(; itChildren != mChildren.end(); ++itChildren)
		(*itChildren)->draw(texture, bDrawTranslucent, u_factor, v_factor);

	glPopMatrix();
}

void FF7Bone::drawWireframe() {
	std::vector<FF7Poly>::iterator it = mPolygons.begin(), itEnd = mPolygons.end();

	glPushMatrix();

	double m[16];

	mOrientation.ToMatrix(m);

	glMultMatrixd(m);
	glPushMatrix();
	glScaled(mScale, mScale, mScale);

	for(; it!=itEnd; ++it) {
		FF7Poly& p = *it;

		switch(p.type) {
		case 0x30:
			glColor4f(1,1,1,1);
			glBegin(GL_TRIANGLES);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			glEnd();
			break;

		case 0x32:
			glColor4f(1,1,1,0.5f);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			glBegin(GL_TRIANGLES);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			glEnd();
			glDisable(GL_BLEND);
			break;

		case 0x38:
			glColor4f(1,1,1,1);
			glBegin(GL_TRIANGLE_STRIP);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			PolyVert(p, 3);
			glEnd();
			break;
		case 0x78:
			glColor4f(1,1,1,1);
			glBegin(GL_TRIANGLES);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			glEnd();
			break;
		}
	}

	glColor3f(1.0f, 1.0f, 0.0f);

	glDisable(GL_FOG);
	glBegin(GL_LINES);
	glVertex3f(0.f, 0.f, 0.f);
	glVertex3f(0.f, 0.f, mLength);
	glEnd();
	glEnable(GL_FOG);

	glPopMatrix();

	glTranslated(0, 0, mLength * mScale);

	std::set<FF7Bone *>::iterator itChildren = mChildren.begin();

	for(; itChildren != mChildren.end(); ++itChildren)
		(*itChildren)->drawWireframe();

	glPopMatrix();
}

void FF7Bone::drawShadow() {
	std::vector<FF7Poly>::iterator it = mPolygons.begin(), itEnd = mPolygons.end();

	glPushMatrix();

	double m[16];

	mOrientation.ToMatrix(m);

	glMultMatrixd(m);
	glPushMatrix();
	glScaled(mScale, mScale, mScale);

	for(; it!=itEnd; ++it) {
		FF7Poly& p = *it;

		switch(p.type) {
		case 0x30:
		case 0x32:
		case 0x78:
			glBegin(GL_TRIANGLES);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			glEnd();
			break;

		case 0x38:
			glBegin(GL_TRIANGLE_STRIP);
			PolyVert(p, 0);
			PolyVert(p, 1);
			PolyVert(p, 2);
			PolyVert(p, 3);
			glEnd();
			break;
		}
	}

	glPopMatrix();
	glTranslated(0, 0, mLength * mScale);

	std::set<FF7Bone *>::iterator itChildren = mChildren.begin();

	for(; itChildren != mChildren.end(); ++itChildren)
		(*itChildren)->drawShadow();

	glPopMatrix();
}

void FF7Bone::enumerate(std::vector<FF7Bone *>& v) {
	v.push_back(this);

	std::set<FF7Bone *>::iterator itChildren = mChildren.begin();

	for(; itChildren != mChildren.end(); ++itChildren)
		(*itChildren)->enumerate(v);
}

void FF7Bone::writeTo(FILE *f) {
	fprintf(f, "%.6f %.6f %.6f %.6f %s\n", mOrientation.xyz.x, mOrientation.xyz.y, mOrientation.xyz.z, mOrientation.w, mName.c_str());
}


void FF7Bone::writeDXF(const char *pszFile) {
	FILE *f = fopen(pszFile, "w");

	fputs(
		"999\n"
		"DXF created from Aeris\n"
		"0\n"
		"SECTION\n"
		"2\n"
		"HEADER\n"
		"9\n"
		"$ACADVER\n"
		"1\n"
		"AC1006\n"
		"9\n"
		"$INSBASE\n"
		"10\n"
		"0.0\n"
		"20\n"
		"0.0\n"
		"30\n"
		"0.0\n"
		"9\n"
		"$EXTMIN\n"
		"10\n"
		"0.0\n"
		"20\n"
		"0.0\n"
		"9\n"
		"$EXTMAX\n"
		"10\n"
		"1000.0\n"
		"20\n"
		"1000.0\n"
		"0\n"
		"ENDSEC\n", f);

	fputs(
		"0\n"
		"SECTION\n"
		"2\n"
		"TABLES\n"
		"0\n"
		"TABLE\n"
		"2\n"
		"LTYPE\n"
		"70\n"
		"1\n"
		"0\n"
		"LTYPE\n"
		"2\n"
		"CONTINUOUS\n"
		"70\n"
		"64\n"
		"3\n"
		"Solid line\n"
		"72\n"
		"65\n"
		"73\n"
		"0\n"
		"40\n"
		"0.000000\n"
		"0\n"
		"ENDTAB\n", f);

	fputs(
		"0\n"
		"TABLE\n"
		"2\n"
		"LAYER\n"
		"70\n"
		"6\n"
		"0\n"
		"LAYER\n"
		"2\n"
		"1\n"
		"70\n"
		"64\n"
		"62\n"
		"7\n"
		"6\n"
		"CONTINUOUS\n"
		"0\n"
		"LAYER\n"
		"2\n"
		"2\n"
		"70\n"
		"64\n"
		"62\n"
		"7\n"
		"6\n"
		"CONTINUOUS\n"
		"0\n"
		"ENDTAB\n"
		"0\n"
		"TABLE\n"
		"2\n"
		"STYLE\n"
		"70\n"
		"0\n"
		"0\n"
		"ENDTAB\n"
		"0\n"
		"ENDSEC\n", f);

	fputs(
		"0\n"
		"SECTION\n"
		"2\n"
		"BLOCKS\n"
		"0\n"
		"ENDSEC\n", f);

	fputs(
		"0\n"
		"SECTION\n"
		"2\n"
		"ENTITIES\n", f);

	std::vector<FF7Poly>::iterator it = mPolygons.begin(), itEnd = mPolygons.end();

	for(; it!=itEnd; ++it) {
		FF7Poly& p = *it;

		switch(p.type) {
		case 0x30:
		case 0x32:
		case 0x78:
			fprintf(f,
				"0\n"
				"3DFACE\n"
				"8\n"
				"1\n"
				"62\n"
				"3\n"
				"10\n"
				"%d\n"
				"20\n"
				"%d\n"
				"30\n"
				"%d\n"
				"11\n"
				"%d\n"
				"21\n"
				"%d\n"
				"31\n"
				"%d\n"
				"12\n"
				"%d\n"
				"22\n"
				"%d\n"
				"32\n"
				"%d\n"
				"13\n"
				"%d\n"
				"23\n"
				"%d\n"
				"33\n"
				"%d\n"
				, mPoints[p.idx[0]>>3].x
				, mPoints[p.idx[0]>>3].y
				, mPoints[p.idx[0]>>3].z
				, mPoints[p.idx[1]>>3].x
				, mPoints[p.idx[1]>>3].y
				, mPoints[p.idx[1]>>3].z
				, mPoints[p.idx[2]>>3].x
				, mPoints[p.idx[2]>>3].y
				, mPoints[p.idx[2]>>3].z
				, mPoints[p.idx[2]>>3].x
				, mPoints[p.idx[2]>>3].y
				, mPoints[p.idx[2]>>3].z);

			break;

		case 0x38:
			fprintf(f,
				"0\n"
				"3DFACE\n"
				"8\n"
				"1\n"
				"62\n"
				"3\n"
				"10\n"
				"%d\n"
				"20\n"
				"%d\n"
				"30\n"
				"%d\n"
				"11\n"
				"%d\n"
				"21\n"
				"%d\n"
				"31\n"
				"%d\n"
				"12\n"
				"%d\n"
				"22\n"
				"%d\n"
				"32\n"
				"%d\n"
				"13\n"
				"%d\n"
				"23\n"
				"%d\n"
				"33\n"
				"%d\n"
				, mPoints[p.idx[0]>>3].x
				, mPoints[p.idx[0]>>3].y
				, mPoints[p.idx[0]>>3].z
				, mPoints[p.idx[1]>>3].x
				, mPoints[p.idx[1]>>3].y
				, mPoints[p.idx[1]>>3].z
				, mPoints[p.idx[3]>>3].x
				, mPoints[p.idx[3]>>3].y
				, mPoints[p.idx[3]>>3].z
				, mPoints[p.idx[2]>>3].x
				, mPoints[p.idx[2]>>3].y
				, mPoints[p.idx[2]>>3].z);

			break;
		}
	}


	fputs(
		"0\n"
		"ENDSEC\n", f);

	fputs(
		"0\n"
		"EOF\n", f);

	fclose(f);
}

static void TransformPoint(float *dst, const FF7Point& pt, const float *transform) {
	double tx, ty, tz, tw;

	double x = pt.x;
	double y = -pt.z;
	double z = pt.y;

	tx = x*transform[0] + y*transform[4] + z*transform[8] + transform[12];
	ty = x*transform[1] + y*transform[5] + z*transform[9] + transform[13];
	tz = x*transform[2] + y*transform[6] + z*transform[10] + transform[14];
	tw = x*transform[3] + y*transform[7] + z*transform[11] + transform[15];

	double rhw = 1.0 / tw;

	dst[0] = (float)(tx * rhw);
	dst[1] = (float)(ty * rhw);
	dst[2] = (float)(tz * rhw);
}

void FF7Bone::writeDXFFromOpenGLTransform(DXF3DOut& ofile) {
	float transform[16];

	glGetFloatv(GL_MODELVIEW_MATRIX, transform);

	std::vector<FF7Poly>::iterator it = mPolygons.begin(), itEnd = mPolygons.end();
	float dst[12];

	for(; it!=itEnd; ++it) {
		FF7Poly& p = *it;

		switch(p.type) {
		case 0x30:
		case 0x32:
		case 0x78:
			TransformPoint(dst+0, mPoints[p.idx[0]>>3], transform);
			TransformPoint(dst+3, mPoints[p.idx[1]>>3], transform);
			TransformPoint(dst+6, mPoints[p.idx[2]>>3], transform);

			ofile.writeTriangle(dst);
			break;

		case 0x38:
			TransformPoint(dst+0, mPoints[p.idx[0]>>3], transform);
			TransformPoint(dst+3, mPoints[p.idx[1]>>3], transform);
			TransformPoint(dst+6, mPoints[p.idx[3]>>3], transform);
			TransformPoint(dst+9, mPoints[p.idx[2]>>3], transform);

			ofile.writeQuad(dst);
			break;
		}
	}
}

void FF7Bone::attachTo(FF7Bone& parent, double length) {
	parent.mChildren.insert(this);

	mpParentNode = &parent;
	mLength = length;
}

void FF7Bone::PolyVert(const FF7Poly& p, int v) {
	FF7Point& pt = mPoints[p.idx[v]>>3];

	glVertex3i(pt.x, pt.y, pt.z);
}

void FF7Bone::PolyVertColor(const FF7Poly& p, int v) {
	FF7Point& pt = mPoints[p.idx[v]>>3];

	glColor3f(p.c[v][0]/255.0f, p.c[v][1]/255.0f, p.c[v][2]/255.0f);
	glVertex3i(pt.x, pt.y, pt.z);
}

void FF7Bone::PolyVertColorTrans(const FF7Poly& p, int v) {
	FF7Point& pt = mPoints[p.idx[v]>>3];

	glColor4f(p.c[v][0]/255.0f, p.c[v][1]/255.0f, p.c[v][2]/255.0f, 0.5f);
	glVertex3i(pt.x, pt.y, pt.z);
}

void FF7Bone::_setupTransform(bool bFinal) {
	if (mpParentNode)
		mpParentNode->_setupTransform(false);

	double m[16];

	mOrientation.ToMatrix(m);
	glMultMatrixd(m);

	if (!bFinal)
		glTranslated(0, 0, mLength);
}

void FF7Bone::setupTransform() {
	_setupTransform(true);
}

Quaternion FF7Bone::getPreOrientation() {
	return mpParentNode ? mpParentNode->getOrientation() : Quaternion(Vector3(0,0,0),1);
}

Quaternion FF7Bone::getOrientation() {
	return mpParentNode ? mpParentNode->getOrientation() * mOrientation : mOrientation;
}

Vector3 FF7Bone::getPosition() {
	Vector3 pt(0,0,0);
	FF7Bone *pBone = mpParentNode;

	while(pBone) {
		pt.z += pBone->mLength;
		pt = pBone->mOrientation.RotatePoint(pt);
		pBone = pBone->mpParentNode;
	}

	return pt;
}

///////////////////////////////

class FF7Pack {
public:
	FF7Pack(const void *src) : mpsrc(static_cast<const unsigned *>(src)) {}
	~FF7Pack() throw() {}

	int size() const { return mpsrc[0]; }

	const FF7PackEntry operator[](int i) {
		FF7PackEntry e = { (const char *)mpsrc + mpsrc[1+i], mpsrc[2+i] - mpsrc[1+i] };

		return e;
	}

private:
	const unsigned *mpsrc;
};

class FF7ModelPack {
public:
	FF7ModelPack(const void *src, int len) : mpsrc(static_cast<const PackHeader *>(src)), mlen(len) {}
	~FF7ModelPack() throw() {}

	int size() const { return mpsrc->count; }

	const FF7BoneEntry operator[](int i) {
		FF7BoneEntry e = {
			(const char *)mpsrc + mpsrc->parts[i].offset,
			0,
			mpsrc->parts[i].parent,
			(signed short)mpsrc->parts[i].bonelength
		};

		if (!mpsrc->parts[i].offset) {
			e.p = NULL;
			return e;
		}

		int j=i;
		while(++j<mpsrc->count)
			if (mpsrc->parts[j].offset) {
				e.len = mpsrc->parts[j].offset - mpsrc->parts[i].offset;
				return e;
			}

		e.len = mlen - mpsrc->parts[i].offset;

		return e;
	}

private:
	const struct PackHeader {
		unsigned short count;
		unsigned short pad[5];
		struct {
			unsigned short parent;		// bone parent ID (0 == root)
			signed short bonelength;	// length of bone
			unsigned short offset;		// offset to geometry
			unsigned short something1;	// always zero
		} parts[1];
	} *mpsrc;
	const int mlen;
};

class MemoryFile {
public:
	void init(const char *fname) {
		FILE *f = fopen(fname, "rb");

		if (f) {
			fseek(f, 0, SEEK_END);
			mBuffer.resize(ftell(f));
			fseek(f, 0, SEEK_SET);
			fread(&mBuffer[0], mBuffer.size(), 1, f);
			fclose(f);
		}
	}

	int size() const { return mBuffer.size(); }

	operator const void *() const { return &mBuffer[0]; }

private:
	std::vector<char> mBuffer;
};

///////////////////////////////

class FF7Model {
public:
	FF7Model();
	~FF7Model();

	void readFrom(FF7Pack& pack, int texture_id);

	void draw();
	void drawWireframe();
	void drawShadow();

	void readFrom(FILE *f);
	void writeTo(FILE *f);

	void addBone(FF7Pack& pack, int model_packet, int attach_point, double length, const char *pszName);

	void setTextures(const GLuint *textures);

	int getBoneCount() const { return mBones.size(); }
	FF7Bone& getBone(int b) { return mBones[b]; }

	Vector3 getPosition() const { return mPosition; }
	void setPosition(const Vector3& p) { mPosition = p; }

	std::string getName() const { return mName; }
	void setName(const std::string& name) { mName = name; }

	void setupTransform();
	void loadPose(FF7Pack& pack, int pose_packet);
private:
	std::deque<FF7Bone> mBones;
	Vector3 mPosition;

	std::vector<GLuint> mTextures;
	double mUFactor, mVFactor;

	std::string mName;
};

FF7Model::FF7Model()
	: mPosition(0,0,0)
{
}

FF7Model::~FF7Model() {
}

void FF7Model::readFrom(FF7Pack& pack, int texture_id) {
	FF7ModelPack modelpack(pack[0].p, pack[0].len);
	int nBones = modelpack.size() + 1;

	mBones.resize(nBones);

	mBones[0].setName("root");
	mBones[0].setLength(0);

	for(int b=1; b<nBones; ++b) {
		FF7BoneEntry ent(modelpack[b-1]);

		if (ent.p)
			mBones[b].readFrom(ent);

		mBones[b].attachTo(mBones[ent.parent_bone], ent.length);
	}

	int w2, h2;

	LoadTextures(mTextures, pack[texture_id], w2, h2);

	mUFactor = 1.0 / w2;
	mVFactor = 1.0 / h2;
}

void FF7Model::draw() {
	glPushMatrix();
	glTranslated(mPosition.x, mPosition.y, mPosition.z);
	mBones[0].draw(&mTextures[0], false, mUFactor, mVFactor);
	glEnable(GL_POLYGON_OFFSET_FILL);
	glPolygonOffset(0,-1);
	mBones[0].draw(&mTextures[0], true, mUFactor, mVFactor);
	glPolygonOffset(0,0);
	glDisable(GL_POLYGON_OFFSET_FILL);
	glPopMatrix();
}

void FF7Model::drawWireframe() {
	glPushMatrix();
	glTranslated(mPosition.x, mPosition.y, mPosition.z);
	mBones[0].drawWireframe();
	glPopMatrix();
}

void FF7Model::drawShadow() {
	glPushMatrix();
	glTranslated(mPosition.x, mPosition.y, mPosition.z);
	mBones[0].drawShadow();
	glPopMatrix();
}

void FF7Model::readFrom(FILE *f) {
	int b;
	int bones = mBones.size();
	char buf[1024], bonename[128];

	while(fgets(buf, sizeof buf, f)) {
		Quaternion q;
		int gotten = sscanf(buf, "%lf %lf %lf %lf %s \n", &q.xyz.x, &q.xyz.y, &q.xyz.z, &q.w, bonename);

		if (gotten >= 5) {
			for(b=0; b<bones; ++b)
				if (!stricmp(bonename, mBones[b].getName().c_str())) {
					mBones[b].setLocalOrientation(q);
					break;
				}
		}
	}
}

void FF7Model::writeTo(FILE *f) {
	int bones = mBones.size();

	for(int b=0; b<bones; ++b) {
		mBones[b].writeTo(f);
	}
}

void FF7Model::addBone(FF7Pack& pack, int model_packet, int attach_point, double length, const char *pszName) {
	int nBones = mBones.size();

	mBones.push_back(FF7Bone());

	mBones[nBones].readFrom((const char *)pack[model_packet].p + 12, pack[model_packet].len - 12);
	mBones[nBones].setName(pszName);
	mBones[nBones].attachTo(mBones[attach_point], length);
}

void FF7Model::setupTransform() {
	glTranslated(mPosition.x, mPosition.y, mPosition.z);
}

void FF7Model::setTextures(const GLuint *textures) {
	mTextures[0] = textures[0];
	mTextures[1] = textures[1];
}

class twelvebit_iterator {
protected:
	const unsigned char *src;
	bool half;
public:
	twelvebit_iterator(const void *s) : src((const unsigned char *)s), half(false) {}

	int operator*() const {
		int v = (src[0]<<8) + src[1];

		return half ? v&0xfff : v>>4;
	}

	twelvebit_iterator& operator++() {
		++src;
		if (half)
			++src;

		half = !half;
		return *this;
	}

	twelvebit_iterator operator++(int) {
		twelvebit_iterator it(*this);

		++*this;

		return it;
	}
};

void FF7Model::loadPose(FF7Pack& pack, int pose_packet) {
	const unsigned char *src = (const unsigned char *)pack[pose_packet].p + 11;
	int nBones = mBones.size();
	int skipmask = 0;

	if (0 && src[-1]) {
		src += 4;
		skipmask = 0x00000C2 * 2 + 1;
	}

	twelvebit_iterator it(src);

	for(int i=0; i<nBones; ++i) {
		if (!(skipmask & 0x01)) {
			int x = *it++;
			int y = *it++;
			int z = *it++;

			mBones[i].setOrientation(Quaternion(Vector3(0,0,0), 1));
			mBones[i].rotateXYZ(x, y, z);
		}
		skipmask >>= 1;
	}
}

///////////////////////////////


MemoryFile file;
MemoryFile file2;

std::vector<FF7Model> g_models;

///////////////////////////////

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    MSG        msg;
    WNDCLASS   wndclass;

	file2.init("earith2.bin");
	file.init("cloud.bin");

	FF7Pack pack1(file);
	FF7ModelPack pack2(pack1[0].p, pack1[0].len);

	if(false) for(int i=0; i<pack2.size(); ++i) {
		FF7Bone m;

		if (pack2[i].p) {
			char buf[128];

			sprintf(buf, "f:/ff7/exploded/cloud%d.dxf", i);

			m.readFrom(pack2[i]);
			m.writeDXF(buf);
		}
	}

//	model.readFrom((const char *)pack1[109].p + 12, pack1[109].len - 12);
//	model0.readFrom(pack2[0]);
//	model2.readFrom(pack2[19]);
//	model2.readFrom((const char *)pack1[1].p + 12, pack1[1].len-12);

    /* Register the frame class */
    wndclass.style         = CS_OWNDC|CS_HREDRAW|CS_VREDRAW;
    wndclass.lpfnWndProc   = (WNDPROC)MainWndProc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = 0;
    wndclass.hInstance     = hInstance;
    wndclass.hIcon         = LoadIcon (NULL, IDI_APPLICATION);
    wndclass.hCursor       = LoadCursor (NULL,IDC_ARROW);
    wndclass.hbrBackground = NULL;
    wndclass.lpszMenuName  = NULL;
    wndclass.lpszClassName = szAppName;

    if (!RegisterClass (&wndclass) )
        return FALSE;

    /* Create the frame */
    g_hwnd = CreateWindow (szAppName,
             "GLAeris",
         WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
//			WS_POPUP | WS_VISIBLE,
             CW_USEDEFAULT,
             CW_USEDEFAULT,
             640,
             480,
             NULL,
             NULL,
             hInstance,
             NULL);

    /* make sure window was created */
    if (!g_hwnd)
        return FALSE;

    /* show and update main window */
    ShowWindow (g_hwnd, nCmdShow);

    UpdateWindow (g_hwnd);


    /* animation loop */
    while (1) {
        /*
         *  Process all pending messages
         */

        while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE) == TRUE)
        {
            if (GetMessage(&msg, NULL, 0, 0) )
            {
                TranslateMessage(&msg);
                DispatchMessage(&msg);
            } else {
                return TRUE;
            }
        }

        drawScene();
    }
}

float cursorx, cursory, cursorz;
static double xrot = 0.0f;
static double yrot = 0.0f;
static double xwrot = 0.0f;
static double ywrot = 180.0f;
static double xzrot = 0.0f;
static double yzrot = 0.0f;

static double xrot2 = -117.0; //-46.0;	//0.0f;
static double yrot2 = -23.0; //-60.0; //0.0f;
static double xwrot2 = -40.0; //-45.0; //0.0f;
static double ywrot2 = 113.0; //116.0; //180.0f;
static double xzrot2 = 87.0; //87.0; //0.0f;
static double yzrot2 = 36.0; //32.0; //0.0f;

double camera_theta = 0, camera_phi = 0;

bool g_bCursor;
int g_nJoint;
int g_nCurrentModel = 0;
FF7Bone *g_pCurrentBone;
FF7Model *g_pCurrentModel;
int g_nAxis;
bool g_bWireframe;
bool g_bTranslate;
bool g_bAdvancedShadow;
bool g_bAntialiased;

double g_camera_zpos = 500;


double g_clipPlaneHeading = 0;
double g_clipPlanePitch = 0;
Vector3 g_clipPlaneOrigin(0,0,-400.0);
bool g_bClipPlane;

FF7Bone *g_IKFirst;
FF7Bone *g_IKSecond;
FF7Bone *g_IKThird;
Quaternion g_IKFirstOrient;
Quaternion g_IKSecondOrient;
Quaternion g_IKThirdOrient;
Vector3 g_IKRotationVector;
double g_IKDistance;

char g_debug[1024];

static int goodmod(int x, int y) {
	x %= y;
	return x<0 ? x+y : x;
}

LONG WINAPI MainWndProc (HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	static bool sDragging = false;
	static bool sDragCamera;
	static int xorigin, yorigin;
	static char buf[256];
		
    LONG    lRet = 1;
    PAINTSTRUCT    ps;
    RECT rect;

    switch (uMsg) {

    case WM_CREATE:
        ghDC = GetDC(hWnd);
        if (!bSetupPixelFormat(ghDC))
            PostQuitMessage (0);

        ghRC = wglCreateContext(ghDC);
        wglMakeCurrent(ghDC, ghRC);
		OutputDebugString((char *)glGetString(GL_RENDERER));
        GetClientRect(hWnd, &rect);
        initializeGL(rect.right, rect.bottom);

		g_pCurrentBone = &g_pCurrentModel->getBone(0);
        break;

    case WM_PAINT:
        BeginPaint(hWnd, &ps);
        EndPaint(hWnd, &ps);
        break;

    case WM_SIZE:
        GetClientRect(hWnd, &rect);
        resize(rect.right, rect.bottom);
        break;

    case WM_CLOSE:
        if (ghRC)
            wglDeleteContext(ghRC);
        if (ghDC)
            ReleaseDC(hWnd, ghDC);
        ghRC = 0;
        ghDC = 0;

        DestroyWindow (hWnd);
        break;

    case WM_DESTROY:
        if (ghRC)
            wglDeleteContext(ghRC);
        if (ghDC)
            ReleaseDC(hWnd, ghDC);

        PostQuitMessage(0);
        break;

	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
		if (!sDragging) {
			sDragging = true;
			sDragCamera = (uMsg == WM_RBUTTONDOWN);
			xorigin = LOWORD(lParam);
			yorigin = HIWORD(lParam);
			SetCapture(hWnd);
		}
		break;

	case WM_LBUTTONUP:
	case WM_RBUTTONUP:
		sDragging = false;
		ReleaseCapture();
		break;

	case WM_MOUSEMOVE:
		if (sDragging) {
			int x = (SHORT)LOWORD(lParam);
			int y = (SHORT)HIWORD(lParam);

			if (sDragCamera) {
				if (GetKeyState(VK_LBUTTON)<0) {
					g_camera_zpos += (y - yorigin) * 0.3f;
				} else {
					camera_theta += (x - xorigin) * 0.3f;
					camera_phi += (y - yorigin) * 0.1f;
				}
			} else if (g_bClipPlane) {
				if (GetKeyState(VK_SHIFT)<0) {
					g_clipPlanePitch -= (y-yorigin) * 0.01f;
					g_clipPlaneHeading += (x-xorigin) * 0.03f;
				} else {
					if (GetKeyState(VK_LBUTTON)<0) {
						if (GetKeyState(VK_RBUTTON)<0) {
							g_clipPlaneOrigin.z -= y-yorigin;
						} else {
							g_clipPlaneOrigin.x += x-xorigin;
							g_clipPlaneOrigin.y += y-yorigin;
						}
					}
				}

				xorigin = x;
				yorigin = y;
			} else {
				if (GetKeyState(VK_LBUTTON)<0) {
					if (g_bTranslate) {
						if (GetKeyState(VK_RBUTTON)<0) {
							g_pCurrentModel->setPosition(g_pCurrentModel->getPosition() + Vector3(0, 0, yorigin-y));
						} else {
							g_pCurrentModel->setPosition(g_pCurrentModel->getPosition() + Vector3(x-xorigin, y-yorigin, 0));
						}
					} else {
						if (GetKeyState(VK_RBUTTON)<0) {
							xzrot += (x - xorigin);
							yzrot += (y - yorigin);
						} else {
							switch(g_nAxis) {
							case 0:
								g_pCurrentBone->rotateY(x - xorigin);
								break;
							case 1:
								g_pCurrentBone->rotateX(yorigin - y);
								break;
							case 2:
								g_pCurrentBone->rotateZ(x - xorigin);
								break;
							case 3:
								// Compute a quaternion for rotating around the Y axis, then apply the inverse
								// view transform to bring it into world space, then the inverse bone transform
								// to get it into bone space.
								{
									Quaternion qView(Quaternion(degreesToRadians(camera_phi), Vector3(1,0,0)) * Quaternion(degreesToRadians(camera_theta), Vector3(0,0,-1)));
									Quaternion qBone(g_pCurrentBone->getPreOrientation());
									Quaternion qRotate(degreesToRadians(x-xorigin), Vector3(0,1,0));

									g_pCurrentBone->rotate(~qBone * ~qView * qRotate * qView * qBone);
								}
								break;
							case 4:
								{
									double delta = (yorigin - y);
									double l1 = fabs(g_IKFirst->getLength());
									double l2 = fabs(g_IKSecond->getLength());
									double mn = fabs(l1-l2) + 1;
									double mx = l1+l2 - 1;

									g_IKDistance += delta;

									if (g_IKDistance < mn)
										g_IKDistance = mn;

									if (g_IKDistance > mx)
										g_IKDistance = mx;

									// Compute new quaternion for center joint.
									double l3 = g_IKDistance;

									double costheta = -(l1*l1 + l2*l2 - l3*l3) / (2.0 * l1 * l2);
									double coshalftheta = sqrt(0.5 + 0.5*costheta);
									double sinhalftheta = sqrt(0.5 - 0.5*costheta);
									Quaternion qNew1(g_IKRotationVector * sinhalftheta, coshalftheta);
		//							Quaternion qNew2((g_IKFirstOrient * ~qNew1).slerp(0.5) * g_IKSecondOrient);
		//							Quaternion qNew2((g_IKFirstOrient * ~qNew1) * g_IKSecondOrient);

									Quaternion qTemp(g_IKSecondOrient * ~qNew1);

									qTemp.w = sqrt(0.5 + 0.5*qTemp.w);
									qTemp.xyz = qTemp.xyz.norm() * sqrt(1.0 - qTemp.w*qTemp.w);

									sprintf(g_debug, "angle: %g\n", acos(qTemp.w));

									Quaternion qNew2(g_IKFirstOrient * qTemp);

									g_IKThird->setLocalOrientation(g_IKThirdOrient * qTemp);
									g_IKSecond->setLocalOrientation(qNew1);
									g_IKFirst->setLocalOrientation(qNew2);
								}
								break;
							}
						}
					}
				}
			}

#if 0
			model_aeris_l_leg.setRotationX(yzrot);
			model_aeris_l_leg.setRotationZ(xzrot);
			model_aeris_l_leg_upper.setRotationX(yrot);
			model_aeris_l_leg_upper.setRotationZ(xrot);
#endif

			xorigin = x;
			yorigin = y;
		}

		break;

	case WM_CHAR:
		switch(wParam) {
#if 0
		case 'i':	cursorx -= 1.0f; break;
		case 'k':	cursorx += 1.0f; break;
		case 'j':	cursory -= 1.0f; break;
		case 'l':	cursory += 1.0f; break;
		case 'o':	cursorz -= 1.0f; break;
		case 'p':	cursorz += 1.0f; break;
		case 'c':	g_bCursor = !g_bCursor; break;
#else
		case 'i':
			{
				FF7Bone *pBone0 = g_pCurrentBone;
				FF7Bone *pBone1 = pBone0->getParent();
				if (!pBone1)
					break;
				FF7Bone *pBone2 = pBone1->getParent();
				if (!pBone2)
					break;
				double l1 = pBone1->getLength();
				double l2 = pBone2->getLength();
				Quaternion q(pBone1->getLocalOrientation());

				if (q.w < 0)
					q = -q;

				g_IKFirst = pBone2;
				g_IKSecond = pBone1;
				g_IKThird = pBone0;
				g_IKFirstOrient = pBone2->getLocalOrientation();
				g_IKSecondOrient = pBone1->getLocalOrientation();
				g_IKThirdOrient = pBone0->getLocalOrientation();
				g_IKRotationVector = q.xyz.norm();
				g_IKDistance = l1*l1 + l2*l2 - 2.0*l1*l2*q.w;

				g_nAxis = 4;
			}
			break;
#endif
		case 'w':	g_bWireframe = !g_bWireframe; break;
		case 't':	g_bTranslate = !g_bTranslate; break;
		case 's':	g_bAdvancedShadow = !g_bAdvancedShadow; break;
		case 'a':	g_bAntialiased = !g_bAntialiased; break;
		case 'c':	g_bClipPlane = !g_bClipPlane; break;
		case '[':
			g_nJoint = goodmod(g_nJoint - 1, g_pCurrentModel->getBoneCount());
			g_pCurrentBone = &g_pCurrentModel->getBone(g_nJoint);
			g_nAxis %= 4;
			break;
		case ']':
			g_nJoint = goodmod(g_nJoint + 1, g_pCurrentModel->getBoneCount());
			g_pCurrentBone = &g_pCurrentModel->getBone(g_nJoint);
			g_nAxis %= 4;
			break;
		case '{':
			g_nCurrentModel = goodmod(g_nCurrentModel - 1, g_models.size());
			g_pCurrentModel = &g_models[g_nCurrentModel];
			g_nJoint = 0;
			g_pCurrentBone = &g_pCurrentModel->getBone(0);
			g_nAxis = 0;
			break;
		case '}':
			g_nCurrentModel = goodmod(g_nCurrentModel + 1, g_models.size());
			g_pCurrentModel = &g_models[g_nCurrentModel];
			g_nJoint = 0;
			g_pCurrentBone = &g_pCurrentModel->getBone(0);
			g_nAxis = 0;
			break;
		case ' ':
			g_nAxis = (g_nAxis+1) % 4;
			break;
		}
#if 0
		{
			sprintf(buf, "x=%g, y=%g, z=%g (%04x %04x %04x) [%04x %04x %04x]"
					, cursorx, cursory, cursorz
					, 0xffff&(short)floor(cursorx+0.5)
					, 0xffff&(short)floor(cursory+0.5)
					, 0xffff&(short)floor(cursorz+0.5)
					, 0xffff&~(short)floor(cursorx+0.5)
					, 0xffff&~(short)floor(cursory+0.5)
					, 0xffff&~(short)floor(cursorz+0.5)
					);
			SetWindowText(hWnd, buf);
		}
#endif

		return 0;

	case WM_KEYDOWN:
		if (wParam == VK_F2) {
			std::string fname(g_pCurrentModel->getName() + ".txt");

			if (IDOK == MessageBox(hWnd, ("Save to "+fname+"?").c_str(), "GLAeris", MB_OKCANCEL)) {
				FILE *f = fopen(fname.c_str(), "w");

				if (f)
					g_pCurrentModel->writeTo(f);

				fclose(f);
			}
			return 0;
		} else if (wParam == VK_F3) {
			std::string fname(g_pCurrentModel->getName() + ".txt");

			if (IDOK == MessageBox(hWnd, ("Load from "+fname+"?").c_str(), "GLAeris", MB_OKCANCEL)) {
				FILE *f = fopen(fname.c_str(), "r");

				if (f)
					g_pCurrentModel->readFrom(f);

				fclose(f);
			}
			return 0;
		} else if (wParam == VK_BACK) {
			SetWindowText(hWnd, "Bone orientation reset.");
			g_pCurrentBone->setOrientation(Quaternion(Vector3(0,0,0), 1));
		} else if (wParam == VK_LEFT)
			newidx -= 0x08;
		else if (wParam == VK_RIGHT)
			newidx += 0x08;

		if (newidx<0) newidx=0; else if (newidx>0x1c8) newidx=0x1c8;

		return 0;

    default:
        lRet = DefWindowProc (hWnd, uMsg, wParam, lParam);
        break;
    }

    return lRet;
}

BOOL bSetupPixelFormat(HDC hdc)
{
    PIXELFORMATDESCRIPTOR pfd={0};
    int pixelformat;

    pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
    pfd.nVersion = 1;
    pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
    pfd.iPixelType = PFD_TYPE_RGBA;
    pfd.cColorBits = 16;
    pfd.cDepthBits = 16;
	pfd.cAlphaBits = 0;
    pfd.cAccumBits = 16;
    pfd.cStencilBits = 0;

    if ( (pixelformat = ChoosePixelFormat(hdc, &pfd)) == 0 )
    {
        MessageBox(NULL, "ChoosePixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    if (SetPixelFormat(hdc, pixelformat, &pfd) == FALSE)
    {
        MessageBox(NULL, "SetPixelFormat failed", "Error", MB_OK);
        return FALSE;
    }

    return TRUE;
}

/* OpenGL code */

static GLuint texture[3];
static GLuint g_width, g_height;

GLvoid resize( GLsizei width, GLsizei height )
{
    GLfloat aspect;

    glViewport( 0, 0, width, height );

    aspect = (GLfloat) width / height;

	g_width = width;
	g_height = height;

    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    gluPerspective( 45.0, aspect, 0.5, 200.0 );
    glMatrixMode( GL_MODELVIEW );
}

void Decode4BitTexture(unsigned char *dst, const int pitch, const unsigned char *src, int w, int h, const unsigned short *pal) {
	for(int y=0; y<h; ++y) {
		for(int x=0; x<w; x++) {
			int c = (unsigned char)*src++;

			dst[y*pitch + 8*x + 6] = (pal[c>>4] & 0x001f) << 3;
			dst[y*pitch + 8*x + 5] = (pal[c>>4] & 0x03e0) >> 2;
			dst[y*pitch + 8*x + 4] = (pal[c>>4] & 0x7c00) >> 7;
			dst[y*pitch + 8*x + 7] = (c>>4) ? 0xff : 0;
			dst[y*pitch + 8*x + 2] = (pal[c&15] & 0x001f) << 3;
			dst[y*pitch + 8*x + 1] = (pal[c&15] & 0x03e0) >> 2;
			dst[y*pitch + 8*x + 0] = (pal[c&15] & 0x7c00) >> 7;
			dst[y*pitch + 8*x + 3] = (c&15) ? 0xff : 0;
		}
	}
}

void LoadTextures(std::vector<GLuint>& textures, const FF7PackEntry& ent, int& w2, int& h2) {
	const unsigned char *f = (const unsigned char *)ent.p;

	int nPalettes = *(short *)(f + 0x12);

	// read width and height

	int w = *(short *)(f + nPalettes*0x20 + 0x1c);
	int h = *(short *)(f + nPalettes*0x20 + 0x1e);

	// compute next power of two

	w2=1;
	h2=1;

	while(w2 < w) w2+=w2;
	while(h2 < h) h2+=h2;

	// width was in words, but this is 4-bit

	w *= 2;
	w2 *= 4;

	// create OpenGL textures

	textures.resize(nPalettes);
	glGenTextures(nPalettes, &textures[0]);

	// read in texture a byte at a time and decode

	int pitch = w2*4;

	std::vector<unsigned char> tex(pitch * h2);

	glPixelStorei(GL_UNPACK_ROW_LENGTH, w2);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

	for(int p=0; p<nPalettes; ++p) {
		Decode4BitTexture(&tex[0], pitch, f + 0x20 + 0x20*nPalettes, w, h, (const unsigned short *)(f + 0x14 + 0x20*p));

		// load texture

		glBindTexture(GL_TEXTURE_2D, textures[p]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, w2, h2, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, &tex[0]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}

	if (glGetError())
		__asm int 3
}

GLvoid initializeGL(GLsizei width, GLsizei height)
{
    GLfloat     aspect;

	g_width = width;
	g_height = height;

	SelectObject(ghDC, GetStockObject(DEFAULT_GUI_FONT));
	wglUseFontBitmaps(ghDC, 32, 95, 1032);

    glClearIndex( (GLfloat)BLACK_INDEX);
    glClearDepth( 1.0 );

    glEnable(GL_DEPTH_TEST);
	glDepthMask(GL_TRUE);
//	glEnable(GL_CULL_FACE);

    glMatrixMode( GL_PROJECTION );
    aspect = (GLfloat) width / height;
    gluPerspective( 45.0, aspect, 3.0, 200.0 );
    glMatrixMode( GL_MODELVIEW );

	// load models

	g_models.resize(1);

	if(1){
		FF7Pack packa1(file2);
		FF7Model *pAeris = &g_models[0];

		pAeris->setName("Aeris");

		pAeris->readFrom(packa1, 95);

		// name

		pAeris->getBone(0).setName("root");
		pAeris->getBone(1).setName("hip");
		pAeris->getBone(2).setName("torso");
		pAeris->getBone(3).setName("head");
		pAeris->getBone(4).setName("hair");
		pAeris->getBone(5).setName("hair1");
		pAeris->getBone(6).setName("hair2");
		pAeris->getBone(7).setName("hair3");
		pAeris->getBone(8).setName("hair4");
		pAeris->getBone(9).setName("l_arm1");
		pAeris->getBone(10).setName("l_arm2");
		pAeris->getBone(11).setName("l_arm_upper");
		pAeris->getBone(12).setName("l_arm_lower");
		pAeris->getBone(13).setName("l_hand");
		pAeris->getBone(14).setName("r_arm1");
		pAeris->getBone(15).setName("r_arm2");
		pAeris->getBone(16).setName("r_arm_upper");
		pAeris->getBone(17).setName("r_arm_lower");
		pAeris->getBone(18).setName("r_hand");
		pAeris->getBone(19).setName("l_leg");
		pAeris->getBone(20).setName("l_leg_upper");
		pAeris->getBone(21).setName("l_leg_lower");
		pAeris->getBone(22).setName("l_boot1");
		pAeris->getBone(23).setName("l_boot2");
		pAeris->getBone(24).setName("flap");
		pAeris->getBone(25).setName("flap1");
		pAeris->getBone(26).setName("flap2");
		pAeris->getBone(27).setName("r_leg");
		pAeris->getBone(28).setName("r_leg_upper");
		pAeris->getBone(29).setName("r_leg_lower");
		pAeris->getBone(30).setName("r_boot1");
		pAeris->getBone(31).setName("r_boot2");

#if 0
		pAeris->getBone(1).rotateXYZ(0xC00, 0x153, 0xE15);
		pAeris->getBone(2).rotateXYZ(0xECF, 0x000, 0xFFF);
		pAeris->getBone(3).rotateXYZ(0x1F5, 0x047, 0x055);
		pAeris->getBone(4).rotateXYZ(0xC18, 0x000, 0x000);
		pAeris->getBone(5).rotateXYZ(0xD94, 0xD78, 0x24B);
		pAeris->getBone(6).rotateXYZ(0xF90, 0xE94, 0x056);
		pAeris->getBone(7).rotateXYZ(0x07E, 0x12C, 0x014);
		pAeris->getBone(8).rotateXYZ(0xFC9, 0xFA7, 0x007);
		pAeris->getBone(9).rotateXYZ(0xF0A, 0xE9F, 0x3BD);
		pAeris->getBone(10).rotateXYZ(0xD89, 0xEB8, 0x180);
		pAeris->getBone(11).rotateXYZ(0xCD4, 0x17C, 0x0E8);
		pAeris->getBone(12).rotateXYZ(0xDCD, 0x800, 0x800);
		pAeris->getBone(13).rotateXYZ(0x064, 0xE7B, 0x261);
		pAeris->getBone(14).rotateXYZ(0xF0B, 0x161, 0xC42);
		pAeris->getBone(15).rotateXYZ(0xD89, 0x148, 0xE80);
		pAeris->getBone(16).rotateXYZ(0xD2E, 0x05A, 0xA10);
		pAeris->getBone(17).rotateXYZ(0xF34, 0x000, 0x000);
		pAeris->getBone(18).rotateXYZ(0xE2F, 0xE30, 0x366);
		pAeris->getBone(19).rotateXYZ(0x05D, 0xA30, 0x0B0);
		pAeris->getBone(20).rotateXYZ(0x2EE, 0xDD5, 0x94E);
		pAeris->getBone(21).rotateXYZ(0x07A, 0x000, 0x000);
		pAeris->getBone(22).rotateXYZ(0xE8B, 0xED6, 0x2CE);
		pAeris->getBone(23).rotateXYZ(0xD9B, 0x000, 0x000);
		pAeris->getBone(24).rotateXYZ(0x08A, 0x767, 0x800);
		pAeris->getBone(25).rotateXYZ(0xCD6, 0x000, 0x000);
		pAeris->getBone(26).rotateXYZ(0xFB9, 0x000, 0x000);
		pAeris->getBone(27).rotateXYZ(0x05E, 0x49E, 0x800);
		pAeris->getBone(28).rotateXYZ(0xD04, 0x34C, 0x9A7);
		pAeris->getBone(29).rotateXYZ(0x0E9, 0x000, 0x000);
		pAeris->getBone(30).rotateXYZ(0xE9F, 0x072, 0xEC2);
		pAeris->getBone(31).rotateXYZ(0xD9B, 0x000, 0x000);
#else
		pAeris->loadPose(packa1, 2);		// base pose
//		pAeris->loadPose(packa1, 3);		// critical pose
//		pAeris->loadPose(packa1, 4);
#endif
	}

	if(0) {
		FF7Pack packc1(file);
		FF7Model *pCloud = &g_models[0];

		pCloud->setName("Cloud");

		pCloud->readFrom(packc1, 96);

		// name

		pCloud->getBone(0).setName("root");
		pCloud->getBone(1).setName("hip");
		pCloud->getBone(2).setName("torso");
		pCloud->getBone(3).setName("head");
		pCloud->getBone(4).setName("l_arm1");
		pCloud->getBone(5).setName("l_arm2");
		pCloud->getBone(6).setName("l_arm_upper");
		pCloud->getBone(7).setName("l_arm_lower");
		pCloud->getBone(8).setName("l_hand");
		pCloud->getBone(9).setName("r_arm1");
		pCloud->getBone(10).setName("r_arm2");
		pCloud->getBone(11).setName("r_arm_upper");
		pCloud->getBone(12).setName("r_arm_lower");
		pCloud->getBone(13).setName("r_hand");
		pCloud->getBone(14).setName("l_leg");
		pCloud->getBone(15).setName("l_leg_upper");
		pCloud->getBone(16).setName("l_leg_lower");
		pCloud->getBone(17).setName("l_boot1");
		pCloud->getBone(18).setName("l_boot2");
		pCloud->getBone(19).setName("r_leg");
		pCloud->getBone(20).setName("r_leg_upper");
		pCloud->getBone(21).setName("r_leg_lower");
		pCloud->getBone(22).setName("r_boot1");
		pCloud->getBone(23).setName("r_boot2");

#if 0
		pCloud->getBone(1).rotateXYZ(0xc00, 0xece, 0x000);
		pCloud->getBone(2).rotateXYZ(0xf82, 0xfe5, 0xe8a);
		pCloud->getBone(3).rotateXYZ(0x0bb, 0x0c8, 0x2bc);
		pCloud->getBone(4).rotateXYZ(0xec2, 0xf02, 0xae6);
		pCloud->getBone(5).rotateXYZ(0x186, 0x401, 0x2c1);
		pCloud->getBone(6).rotateXYZ(0x354, 0x1e4, 0xee6);
		pCloud->getBone(7).rotateXYZ(0xe53, 0x000, 0x000);
		pCloud->getBone(8).rotateXYZ(0x02a, 0x061, 0xef1);
		pCloud->getBone(9).rotateXYZ(0xfef, 0x0f0, 0x25f);
		pCloud->getBone(10).rotateXYZ(0x2f1, 0x006, 0x0b5);
		pCloud->getBone(11).rotateXYZ(0x280, 0xac2, 0xe2b);
		pCloud->getBone(12).rotateXYZ(0xf08, 0x000, 0x000);
		pCloud->getBone(13).rotateXYZ(0xfc4, 0xff1, 0xec4);
		pCloud->getBone(14).rotateXYZ(0x000, 0xa2e, 0x000);
		pCloud->getBone(15).rotateXYZ(0x299, 0xdf1, 0xa75);
		pCloud->getBone(16).rotateXYZ(0x11c, 0x000, 0x000);
		pCloud->getBone(17).rotateXYZ(0xd81, 0xff2, 0x308);
		pCloud->getBone(18).rotateXYZ(0xd49, 0xfee, 0x010);
		pCloud->getBone(19).rotateXYZ(0x000, 0x36e, 0x000);
		pCloud->getBone(20).rotateXYZ(0x166, 0xc04, 0x057);
		pCloud->getBone(21).rotateXYZ(0x299, 0x000, 0x000);
		pCloud->getBone(22).rotateXYZ(0xeb9, 0xf87, 0x044);
		pCloud->getBone(23).rotateXYZ(0xd56, 0x000, 0x000);
#else
		pCloud->loadPose(packc1, 2);
#endif
	}

	// manually attach Cloud's weapon to Aeris's right hand

//	g_models[0].addBone(FF7Pack(file), 100, 18, 0, "r_weapon");
//	g_models[0].getBone(32).setScale(0.7);

	g_nCurrentModel = 0;
	g_pCurrentModel = &g_models[g_nCurrentModel];
}

//////

void drawCursor() {
	if (g_bCursor) {
		float x=cursorx, y=cursory, z=cursorz;

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glEnable(GL_BLEND);

		for(int i=0; i<2; ++i) {
			glDepthMask(i?GL_TRUE:GL_FALSE);

			glBegin(GL_QUADS);
			glColor4f(1.f,0.f,0.f,0.25f);
			glVertex3f(-256, -256, z);
			glVertex3f(+256, -256, z);
			glVertex3f(+256, +256, z);
			glVertex3f(-256, +256, z);

			glColor4f(0.f,1.f,0.f,0.25f);
			glVertex3f(-256, y, -256);
			glVertex3f(+256, y, -256);
			glVertex3f(+256, y, +256);
			glVertex3f(-256, y, +256);

			glColor4f(0.f,0.f,1.f,0.25f);
			glVertex3f(x, -256, -256);
			glVertex3f(x, +256, -256);
			glVertex3f(x, +256, +256);
			glVertex3f(x, -256, +256);
			glEnd();
		}
		
		glDisable(GL_BLEND);
	}
}

void MultShadowMatrix(const Vector3& abc, double d, const Vector3& dir) {
	const double matrix[16]={
		abc.y * dir.y + abc.z * dir.z,
		-abc.x * dir.y,
		-abc.x * dir.z,
		0,

		-abc.y * dir.x,
		abc.x*dir.x + abc.z*dir.z,
		-abc.y*dir.z,
		0,

		-abc.z*dir.x,
		-abc.z*dir.y,
		abc.x*dir.x + abc.y*dir.y,
		0,

		-d*dir.x,
		-d*dir.y,
		-d*dir.z,
		abc*dir
	};

	glMultMatrixd(matrix);
}

void drawAxes() {
	glBegin(GL_LINES);
	glColor3f(1,0,0);
	glVertex3i(0,0,0);
	glVertex3i(60,0,0);
	glColor3f(0,1,0);
	glVertex3i(0,0,0);
	glVertex3i(0,60,0);
	glColor3f(0,0,1);
	glVertex3i(0,0,0);
	glVertex3i(0,0,60);
	glEnd();
}

void drawDisc() {
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
	glDepthMask(GL_FALSE);
	glShadeModel(GL_FLAT);
	glBegin(GL_TRIANGLE_FAN);
	{
		glVertex3i(0,0,0);
		for(int d = 1; d<16; ++d) {
			double t = d * (3.14159265358979323846 / 8.0);

			if (d&1)
				glColor4f(1.0f, 1.0f, 0.0f, 0.5f);
			else
				glColor4f(0.8f, 0.8f, 0.0f, 0.5f);
			glVertex2f(60*cos(t), 60*sin(t));
		}
	}
	glEnd();
	glShadeModel(GL_SMOOTH);
	glDepthMask(GL_TRUE);
	glDisable(GL_BLEND);
}

void drawBoneAdjustmentAxis(FF7Bone *pBone, int mode) {
	glPushMatrix();
	pBone->setupTransform();

	drawAxes();

	switch(g_nAxis) {
	case 0:	// Y
		glRotated(90,1,0,0);
		break;
	case 1:	// X
		glRotated(90,0,1,0);
		break;
	case 2: // Z
		break;
	case 3: // screen aligned
		{
			double m[16];

			// Transform the screen aligned disc into world space, then into bone space.

			Quaternion qBone(pBone->getOrientation());
			Quaternion qInvView(Quaternion(degreesToRadians(-camera_theta), Vector3(0,0,-1)) * Quaternion(degreesToRadians(-camera_phi), Vector3(1,0,0)));

			(~qBone * qInvView).ToMatrix(m);
			glMultMatrixd(m);
		}
		glRotated(90,1,0,0);
		break;
	}

	// Draw disc in XY plane.
	drawDisc();
	glPopMatrix();
}

void drawWireframeCharacter() {
	static const float fogcolor[4]={0.2f, 0.5f, 1.0f, 1.0f };

	glFogi(GL_FOG_MODE, GL_LINEAR);
	glFogf(GL_FOG_START, 6.0f);
	glFogf(GL_FOG_END, 14.0f);
	glFogfv(GL_FOG_COLOR, fogcolor);
	glEnable(GL_FOG);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
//	glDepthMask(GL_FALSE);

	int models = g_models.size();

	for(int m=0; m<models; ++m)
		g_models[m].drawWireframe();

//	glDepthMask(GL_TRUE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glDisable(GL_FOG);
}

void drawSolidCharacter() {
	int models = g_models.size();

	for(int m=0; m<models; ++m)
		g_models[m].draw();
}

void drawCharacters() {
	glPushMatrix();
		if (g_bClipPlane) {
			double eqn[4];
			Quaternion orient(Quaternion(g_clipPlaneHeading, Vector3(0,0,-1)) * Quaternion(g_clipPlanePitch, Vector3(1,0,0)));

			Vector3 abc(orient.RotatePoint(Vector3(0,1,0)));

			eqn[0] = abc.x;
			eqn[1] = abc.y;
			eqn[2] = abc.z;
			eqn[3] = -abc * g_clipPlaneOrigin;

			glEnable(GL_CLIP_PLANE0);
			glClipPlane(GL_CLIP_PLANE0, eqn);
			drawWireframeCharacter();

			eqn[0] = -eqn[0];
			eqn[1] = -eqn[1];
			eqn[2] = -eqn[2];
			eqn[3] = -eqn[3];
			glClipPlane(GL_CLIP_PLANE0, eqn);
			drawSolidCharacter();

			glDisable(GL_CLIP_PLANE0);

			if (!g_bAntialiased) {
				glTranslated(g_clipPlaneOrigin.x, g_clipPlaneOrigin.y, g_clipPlaneOrigin.z);

				double m[16];

				orient.ToMatrix(m);

				glMultMatrixd(m);

				drawAxes();
			}
		} else if (g_bWireframe) {
			drawWireframeCharacter();
		} else {
			drawSolidCharacter();
		}
	glPopMatrix();
}

void drawShadow() {
	Vector3 groundNormal(0, 1, 0);
	Vector3 lightDir(1, -1, 2);

	glPushMatrix();

		if (g_bAdvancedShadow) {
#if 1
			Vector3 perpVec(Vector3(lightDir.x, lightDir.z, -lightDir.y) % lightDir);

			glEnable(GL_BLEND);
			glBlendFunc(GL_ONE_MINUS_SRC_ALPHA, GL_SRC_ALPHA);

			for(int i=0; i<128; ++i) {
				glColor4f(0.25f, 0.25f, 0.25f, 0.98f + (0.02f/128.0f) * (128-i));

				Vector3 rotationVec(Quaternion(i*(3.1415926535897932/16.0), lightDir.norm()).RotatePoint(perpVec.norm()));

				Vector3 perturbedLightVec(Quaternion(i*0.0005, rotationVec).RotatePoint(lightDir));

				glPushMatrix();
				MultShadowMatrix(groundNormal, 0, perturbedLightVec.norm());

				glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
				glDepthFunc(GL_ALWAYS);
				{
					int models = g_models.size();

					for(int m=0; m<models; ++m)
						g_models[m].drawShadow();
				}
				glPopMatrix();
				glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

				glDepthFunc(GL_GREATER);
				glDepthMask(GL_FALSE);

				glPushMatrix();
				glLoadIdentity();
				glMatrixMode(GL_PROJECTION);
				glPushMatrix();
				glLoadIdentity();
				glBegin(GL_QUADS);
				glVertex3f(-1.f,-1.f,+0.99f);
				glVertex3f(+1.f,-1.f,+0.99f);
				glVertex3f(+1.f,+1.f,+0.99f);
				glVertex3f(-1.f,+1.f,+0.99f);
				glEnd();
				glPopMatrix();
				glMatrixMode(GL_MODELVIEW);
				glPopMatrix();

				glDepthMask(GL_TRUE);
				glClear(GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
			}

			glDepthFunc(GL_LESS);
			glDisable(GL_BLEND);
#else
			Vector3 perpVec(Vector3(lightDir.x, lightDir.z, -lightDir.y) % lightDir);
			glDisable(GL_DEPTH_TEST);
			glColor3f(0,0,0);

			for(int i=0; i<1; ++i) {
				Vector3 rotationVec(Quaternion(i*(3.1415926535897932/16.0), lightDir.norm()).RotatePoint(perpVec.norm()));

				Vector3 perturbedLightVec(Quaternion(i*0.0005, rotationVec).RotatePoint(lightDir));

				glPushMatrix();
				MultShadowMatrix(groundNormal, 0, perturbedLightVec.norm());

				{
					int models = g_models.size();

					for(int m=0; m<models; ++m)
						g_models[m].drawShadow();
				}
				glPopMatrix();

				glAccum(i?GL_ACCUM:GL_LOAD, 1.0/128.0);

				glClear(GL_COLOR_BUFFER_BIT);
			}

			glAccum(GL_RETURN, 1.0);

			glEnable(GL_DEPTH_TEST);
#endif
		} else {
			glColor3f(0,0,0);

			MultShadowMatrix(groundNormal, 0, lightDir);

			int models = g_models.size();

			for(int m=0; m<models; ++m)
				g_models[m].drawShadow();
		}
	glPopMatrix();
}

static __int64 __declspec(naked) __cdecl rdtsc() {
	__asm {
		push ebx
		xor eax,eax
		cpuid
		pop ebx
		rdtsc
		ret
	}
}

GLuint g_blitTexture;

void drawAntialiasedCharacters() {
	const int pitch = ((g_width+1)&~1);
	const int s = pitch*g_height;
	std::vector<unsigned short> accum(s*4);
	std::vector<unsigned char> image(s*4);

	std::fill(accum.begin(), accum.end(), 0);

	glPixelStorei(GL_PACK_ALIGNMENT, 8);
	glPixelStorei(GL_PACK_ROW_LENGTH, 0);
	glPixelStorei(GL_UNPACK_ALIGNMENT, 8);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);

	double projmatrix[16];

	glMatrixMode(GL_PROJECTION);

	glGetDoublev(GL_PROJECTION_MATRIX, projmatrix);

	__int64 readcycles = 0, writecycles = 0;

static const int AA_BITS = 3;
static const int AA_LEVEL = 1<<AA_BITS;

	for(int yo=0; yo<AA_LEVEL; ++yo)
		for(int xo=0; xo<AA_LEVEL; ++xo) {
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

			glLoadIdentity();
			glTranslated((2*xo - (AA_LEVEL-1)) * (1.0 / (AA_LEVEL*g_width)), (2*yo - (AA_LEVEL-1)) * (1.0 / (AA_LEVEL*g_height)), 0);
			glMultMatrixd(projmatrix);

			glMatrixMode(GL_MODELVIEW);
			drawShadow();
			drawCharacters();
			glMatrixMode(GL_PROJECTION);

			__int64 t1 = rdtsc();
			glReadPixels(0, 0, g_width, g_height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, &image[0]);
			readcycles += rdtsc() - t1;

#if 0
			for(int i=0; i<s*4; ++i) {
				accum[i] += image[i];
			}
#else
			unsigned short *dst = &accum[0];
			const unsigned char *src = &image[0];
			const int count = s;

			__asm {
				mov eax, count
				mov ecx, src
				mov edx, dst
				pxor mm7,mm7
				lea ecx, [ecx + eax*4]
				lea edx, [edx + eax*8]
				neg eax
			sumloop:
				movd mm0, [ecx+eax*4]
				punpcklbw mm0,mm7
				paddw mm0, [edx+eax*8]
				movq [edx+eax*8],mm0
				inc eax
				jne sumloop
				emms
			}
#endif
		}

	glLoadIdentity();
	glOrtho(0, g_width, 0, g_height, -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glRasterPos2i(-1, -1);

	for(int j=0; j<s*4; ++j)
		image[j] = (accum[j] + (AA_LEVEL*AA_LEVEL>>1)) >> (2*AA_BITS);

	glClear(GL_COLOR_BUFFER_BIT);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_ALPHA_TEST);
	glDisable(GL_STENCIL_TEST);
	glDisable(GL_DITHER);
	glDepthMask(GL_FALSE);
	__int64 t2 = rdtsc();


//	glDrawPixels(g_width, g_height, GL_BGRA_EXT, GL_UNSIGNED_BYTE, &image[0]);

	if (!g_blitTexture) {
		glGenTextures(1, &g_blitTexture);
		glBindTexture(GL_TEXTURE_2D, g_blitTexture);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, 256, 256, 0, GL_BGRA_EXT, GL_UNSIGNED_BYTE, &std::vector<long>(256*256)[0]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	}

	glEnable(GL_TEXTURE_2D);
	glPixelStorei(GL_UNPACK_ROW_LENGTH, g_width);
	glBindTexture(GL_TEXTURE_2D, g_blitTexture);
	glColor4d(1.0,1.0,1.0,1.0);
	for(int yp=0; yp<g_height; yp+=256) {
		for(int xp=0; xp<g_width; xp+=256) {
			const int w = std::min<int>(g_width,xp+256)-xp;
			const int h = std::min<int>(g_height,yp+256)-yp;

			glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, w, h, GL_BGRA_EXT, GL_UNSIGNED_BYTE, &image[(pitch*yp+xp)*4]);
			glBegin(GL_TRIANGLE_STRIP);
			glTexCoord2i(0,0);	glVertex2i(xp, yp);
			glTexCoord2i(1,0);	glVertex2i(xp+256, yp);
			glTexCoord2i(0,1);	glVertex2i(xp, yp+256);
			glTexCoord2i(1,1);	glVertex2i(xp+256, yp+256);
			glEnd();
		}
	}

	writecycles += rdtsc() - t2;

	glDepthMask(GL_TRUE);
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixd(projmatrix);
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	{
		char buf[256];
		sprintf(buf, "read %I64d ms, write %I64d ms\n", readcycles / 733000, writecycles / 733000);
		SetWindowText(g_hwnd, buf);
	}
}

GLvoid drawScene()
{
	static const float mapaxes[16]={
		1.f, 0.f, 0.f, 0.f,
		0.f, -1.f, 0.f, 0.f,
		0.f, 0.f, -1.f, 0.f,
		0.f, 0.f, 0.f, 1.f,
	};

//	glClearColor(1,1,1,1);
	glClearColor(0.5,0.5,0.5,1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glShadeModel(GL_SMOOTH);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -10.0f);
	float scale = 1.0f / 128.0f;
	glScalef(scale, scale, scale);

	glRotatef(camera_phi, 1, 0, 0);
	glRotatef(camera_theta, 0, 1, 0);
	glMultMatrixf(mapaxes);
	glTranslated(0, 0, g_camera_zpos);

	glColor3f(1.0f,1.0f,1.0f);

	glAlphaFunc(GL_GREATER, 0);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
	glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

	////////////

	int i;

	if (!g_bAntialiased) {
		drawShadow();
		drawCharacters();
	} else {
		drawAntialiasedCharacters();
		glMatrixMode(GL_MODELVIEW);
	    SWAPBUFFERS;
		return;
	}

#if 0
	glBegin(GL_POINTS);

	for(i=0; i<8; ++i) {
		glVertex3f(i&1?-256:256, i&2?-256:256, i&4?-256:256);
	}

	glEnd();
#endif


	///////////////////////////////////////////

	glPushMatrix();

	g_pCurrentModel->setupTransform();

	if (g_nAxis == 4) {
		FF7Bone *pBone0 = g_pCurrentBone;
		FF7Bone *pBone1 = pBone0->getParent();
		FF7Bone *pBone2 = pBone1->getParent();

		Vector3 v1(0,0,-1);
		Vector3 v2(g_IKRotationVector);
		Vector3 v3(v1 % v2);

		glPushMatrix();
			pBone1->setupTransform();

			glRotated(90, v3.x, v3.y, v3.z);
			drawAxes();
			drawDisc();
		glPopMatrix();

		glPushMatrix();
			pBone2->setupTransform();

			glRotated(90, v3.x, v3.y, v3.z);
			drawAxes();
			drawDisc();
		glPopMatrix();
	} else
		drawBoneAdjustmentAxis(g_pCurrentBone, g_nAxis);

	glPopMatrix();

	drawAxes();

	////////////

	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
		glLoadIdentity();
		glOrtho(0, g_width, 0, g_height, 0.0f, 1.0f);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslated(64, 64, 0);
		glScaled(48, 48, 0);
		glRotatef(camera_phi, 1, 0, 0);
		glRotatef(camera_theta, 0, 1, 0);
		glMultMatrixf(mapaxes);

		glBegin(GL_LINES);
		glColor3f(1,0,0);
		glVertex3i(0,0,0);
		glVertex3i(1,0,0);
		glColor3f(0,1,0);
		glVertex3i(0,0,0);
		glVertex3i(0,1,0);
		glColor3f(0,0,1);
		glVertex3i(0,0,0);
		glVertex3i(0,0,1);
		glEnd();

		glLoadIdentity();

		static char buf[256];
		static const char *const sAxes[]={"heading","pitch","roll","screen-aligned","two-bone IK"};

		glColor3f(0.0f, 0.0f, 0.0f);
		glListBase(1000);

		glRasterPos2i(16, 208);
		sprintf(buf, "Current model: %s", g_pCurrentModel->getName().c_str());
		glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);

		if (g_bTranslate) {
			glRasterPos2i(16, 192);
			Vector3 pos(g_pCurrentModel->getPosition());
			sprintf(buf, "Current position: %.2f, %2.f, %.2f", pos.x, pos.y, pos.z);
			glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
		} else if (g_bClipPlane) {
			glRasterPos2i(16, 192);
			sprintf(buf, "Current clip plane position: %.2f, %.2f, %.2f", g_clipPlaneOrigin.x, g_clipPlaneOrigin.y, g_clipPlaneOrigin.z);
			glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
		} else {
			glRasterPos2i(16, 192);
			sprintf(buf, "Current bone: %s", g_pCurrentBone->getName().c_str());
			glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
			glRasterPos2i(16, 176);
			sprintf(buf, "Current mode: %s", sAxes[g_nAxis]);
			glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
		}

		if (g_nAxis == 4) {
			glRasterPos2i(16, 160);
			sprintf(buf, "IK distance: %.2f", g_IKDistance);
			glCallLists(strlen(buf), GL_UNSIGNED_BYTE, buf);
		}

		glRasterPos2i(16, 144);
		glCallLists(strlen(g_debug), GL_UNSIGNED_BYTE, g_debug);

		glMatrixMode(GL_PROJECTION);

	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	// Flip!

    SWAPBUFFERS;
}


