#include <math.h>

#include "Vector.h"

Vector3 Vector3::operator%(const Vector3& v) const {
	return Vector3(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y-y*v.x);
}

Vector3 Vector3::norm() const {
	double rcp = 1.0/len();

	return Vector3(x*rcp, y*rcp, z*rcp);
}

double Vector3::len() const {
	return sqrt(len2());
}

Vector3 Vector3::rotate2dX(double cs, double sn) const {
	return Vector3(x, z*sn+y*cs, z*cs-y*sn);
}

Vector3 Vector3::rotate2dY(double cs, double sn) const {
	return Vector3(x*cs-z*sn, y, x*sn+z*cs);
}

Vector3 Vector3::rotate2dZ(double cs, double sn) const {
	return Vector3(x*cs-y*sn, x*sn+y*cs, z);
}

///////////////////////////////////////////////////////////////////////////

Quaternion::Quaternion(const Vector3& v1, const Vector3& v2) throw() {
	Vector3 rotaxis(v1 % v2);
	double prodlens_inv = 1.0 / rotaxis.len();
	double cosval = (v1 * v2) * prodlens_inv;

	if (cosval > 1.0)
		cosval = 1.0;
	else if (cosval < -1.0)
		cosval = -1.0;

	w = sqrt(0.5 * (1.0 + cosval));
	xyz = rotaxis * prodlens_inv * sqrt(0.5 * (1.0 - cosval));
}


void Quaternion::ToMatrix(double *dst) const throw() {
	double xy = 2.0 * (xyz.x * xyz.y);
	double xz = 2.0 * (xyz.x * xyz.z);
	double yz = 2.0 * (xyz.y * xyz.z);
	double xw = 2.0 * (xyz.x * w);
	double yw = 2.0 * (xyz.y * w);
	double zw = 2.0 * (xyz.z * w);
	double xx = 2.0 * (xyz.x * xyz.x);
	double yy = 2.0 * (xyz.y * xyz.y);
	double zz = 2.0 * (xyz.z * xyz.z);

	dst[ 0] = 1.0 - (yy+zz);
	dst[ 1] = xy + zw;
	dst[ 2] = xz - yw;
	dst[ 3] = 0.0;

	dst[ 4] = xy - zw;
	dst[ 5] = 1.0 - (xx+zz);
	dst[ 6] = yz + xw;
	dst[ 7] = 0.0;

	dst[ 8] = xz + yw;
	dst[ 9] = yz - xw;
	dst[10] = 1.0 - (xx+yy);
	dst[11] = 0.0;

	dst[12] = 0.0;
	dst[13] = 0.0;
	dst[14] = 0.0;
	dst[15] = 1.0;
}

void Quaternion::Normalize() throw() {
	double norm = xyz.len2() + w*w;
	double inv_norm = 1.0 / sqrt(norm);

	xyz *= inv_norm;
	w *= inv_norm;
}

Vector3 Quaternion::RotatePoint(const Vector3& pt) const throw() {
	Quaternion result((*this) * Quaternion(pt, 1) * ~(*this));

	return result.xyz * (1.0 / result.w);
}

Quaternion Quaternion::slerp(const Quaternion& dst, double t) const throw() {
	double costheta = xyz*dst.xyz + w*w;
	double invsintheta = 1.0 / sqrt(1.0 - costheta*costheta);
	double theta = acos(costheta);

	double c1 = sin(theta*(1.0-t))*invsintheta;
	double c2 = sin(theta*t)*invsintheta;

	Quaternion q(xyz*c1 + dst.xyz*c2, w*c1 + dst.w*c2);

	q.Normalize();

	return q;
}

Quaternion Quaternion::slerp(double t) const throw() {
	return slerp(Quaternion(Vector3(0,0,0),1), 0.5);
}
